///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// File:	hrc4.js
//
// Author:	Copyright (c) 2006,2010 Henk Reints, http://henk-reints.nl
//
// Purpose:	Tool for simple ARCFOUR-based data encryption and other binary operations on String Objects.
//
// Original date:    Jan-2006
// Last modified: 17-Feb-2010
// Update history: see below
//
// Description:
//	This package defines the HRC4 Object, which has methods for converting Strings to/from various MIME formats
//	(hexadecimal, Base64, Quoted Printable) and data encryption based on ARCFOUR principles. The encryption uses
//	a modified initial Key Scheduling Algorithm (KSA) with respect to ARCFOUR, which should (hopefully, I've got
//	no proof) overcome the so called Fluhrer, Mantin and Shamir attack (see http://en.wikipedia.org/wiki/RC4).
//	It accepts keys of ANY length and it uses the entire key, not only the first 256 characters. The number of
//	status initialisation rounds using the key is at least 512 and at least twice the length of the key. After
//	initialisation of the status vector, the PRNG is advanced by 997 rounds before it is available for use.
//	(997 is the largest prime below 1000; I've got no mathematical reason for this number, just think it's nice...)
//
// References:
//	o ARCFOUR: A Stream Cipher Encryption Algorithm "Arcfour"
//	  http://www.mozilla.org/projects/security/pki/nss/draft-kaukonen-cipher-arcfour-03.txt
//	o Wikipedia article on ARCFOUR/RC4: http://en.wikipedia.org/wiki/RC4
//	o RFC2045: Multipurpose Internet Mail Extensions (MIME) Part One: Format of Internet Message Bodies
//	  http://www.rfc-editor.org/rfc/rfc2045.txt
//	o RFC1321: The MD5 Message-Digest Algorithm
//	  http://www.rfc-editor.org/rfc/rfc1321.txt
//	o FIPS PUB 180-2: Secure Hash Standard (SHS)
//	  http://csrc.nist.gov/publications/fips/fips180-2/fips180-2withchangenotice.pdf
//	o RFC3174: US Secure Hash Algorithm 1 (SHA1)
//	  http://www.rfc-editor.org/rfc/rfc3174.txt
//	o RFC2104: HMAC: Keyed-Hashing for Message Authentication,
//	  http://www.rfc-editor.org/rfc/rfc2104.txt
//
// Notes:
//	o The ARCFOUR algorithm is most probably identical to the RC4 algorithm by Ron Rivest (RC4 = Rivest Cypher 4),
//	  which is a registered trademark. It has been reverse-engineered many years ago and the result thereof has
//	  been named ARCFOUR. The HRC4 Object implements ARCFOUR with a modified initialisation algorithm.
//	o Did you know that your own browser most often uses RC4 whenever you go to a secure site using the https
//	  protocol? Whenever you are telebanking? Whenever you are paying anything via Internet with your creditcard
//	  credentials? And did you know that many WIFI versions use RC4?
//	o The HRC4 Object handles String Objects, but interprets their content as a series of BYTES holding values
//	  in the range 0 through 255, where each character code equals the byte value.
//	o In String Objects, text is stored as Unicode characters, i.e. character codes in the range from 0 to 65535.
//	  The first 128 are the good old ASCII table, the range from 160 to 255 is the 'ISO 8859 Latin 1' character
//	  set containing some special symbols and characters with various accents:
//	    ""
//	  The range from 128 to 159 is used by Windows to convert SOME often used special characters to single-byte
//	  values: ""
//	  The values above 255 are for other special symbols and non-Latin based character sets.
//	  The characters from 128 to 159 can be interpreted as bytes, but are internally stored as 2-byte Unicode
//	  characters. This script adds a translation table and 2 methods to the String Object for counteracting that
//	  implicit conversion. Those methods are used by the HRC4 object, the core of this script.
//	o Apart from those few Unicode characters that are converted to the range 128..159, any high order byte of
//	  any charCode is completely ignored by the core cyphering and hashing algorithms in this script, treating
//	  String Objects as BYTE Arrays. For properly handling Unicode Characters above 255 it is necessary to convert
//	  them to other notation, for example by using JavaScript's built-in 'escape' and 'unescape' methods.
//	  The HRC4 Object provides a "soft" escape method, which escapes only non-byte values and, necessarily, the
//	  percent sign. The '**crypt' methods (both static and prototype) of the HRC4 Object already use this to take
//	  care of such conversions, so it can be used to encrypt data in any language.
//	o It is also recommended to use conversion methods for storing or transmitting encyphered data, since it
//	  contains many non-printable characters. Although stored in a String object, encyphered data are to be
//	  considered as a binary stream. The HRC4 Object provides methods for Base64 encoding and decoding, which is
//	  the recommended way to store or transmit binary data. Please note, that, although it makes data unreadable
//	  at first sight, Base64 encoding is _NOT_ encryption!
//	o The HRC4 object contains prototype methods and static methods. Please take good notice of the difference!
//	  Static methods belong to the Constructor of the object, Prototype methods operate on an Object Instance.
//	  To learn more about JavaScript Objects you can for example consult the MSDN Library,
//	  which is found at http://msdn.microsoft.com/library/default.asp, there you navigate as follows:
//	  MSDN Library > Web Development > Scripting > JScript > JScript User's Guide.
//	o The HRC4 Object allows for cyphering multiple consecutive messages which do not use the same key
//	  for every message, but they continue with the internal PRNG status left behind by the last message.
//	  This implies that sender and receiver must both initialise an HRC4 Object with the same key and
//	  then run synchronously to encypher/decypher the messages, as follows:
//		Sender					Receiver
//		init HRC4 object with the key		init HRC4 object with the same key
//		encypher message 1 and transmit		receive message 1 and decypher
//		encypher message 2 and transmit		receive message 2 and decypher
//		etc.					etc.
//	o Please keep in mind that this is JavaScript,
//	  so it is not suitable for operations on very large amounts of data.
//
// Below is a summary of properties and methods, scroll down to the actual code for more details.
//
// Constructor:
//	HRC4(key,arcfourMode)		creates an HRC4 Object Instance
//
// Constructor properties:
//	lastDecypherIsGood		Boolean indicating success/failure of last decypher/decrypt
//					SHOULD BE TREATED AS READ-ONLY!
//
//	_MD5table			Array of Numbers needed for computing the MD5 hash of a message.
//					SHOULD BE TREATED AS INTERNAL DATA OF THE OBJECT, DO NOT MODIFY IT!
//
// Constructor methods (static methods):
//	new1      (key,arcfourMode)	obtains a new HRC4 Object initialised with the old 'init'  prototype method
//	new2      (key1,key2)		obtains a new HRC4 Object initialised with the new 'init2' prototype method
//	escape    (txt)			escapes non-byte Unicode characters
//	toHex     (msg,sep)		converts a String  to  hexadecimal notation
//	fromHex   (msg)			converts a String from hexadecimal notation
//	toDec     (msg,sep)		converts a String  to  decimal notation
//	fromDec   (msg)			converts a String from decimal notation
//	toBase64  (msg,wrap)		converts a String  to  Base64 notation
//	fromBase64(msg)			converts a String from Base64 notation
//	toQuotPr  (msg)			converts a String  to  Quoted Printable notation
//	fromQuotPr(msg)			converts a String from Quoted Printable notation
//	unicodeToBytes(msg)		splits every Unicode character into 2 single bytes
//	bytesToUnicode(msg)		combines every pair of bytes to a single Unicode character
//
//	MD5   (msg)			computes the MD5 hash (Message Digest version 5) of a message
//	SHA1  (msg)			computes the SHA1 hash (Secure Hash Algorithm) of a message
//	SHA256(msg)			computes the SHA256 hash (Secure Hash Algorithm) of a message
//	FP    (msg,size)		my own variant for computing a fingerprint using the HRC4 KSA + PRNG
//	FP32  (msg)			my own variant for computing a 32-byte fingerprint using the HRC4 KSA + PRNG
//	FP64  (msg)			my own variant for computing a 64-byte fingerprint using the HRC4 KSA + PRNG
//
//	HMAC(msg,key,hfun,blksz)	generic method for computing a keyed hash
//	HMAC_MD5   (msg,key)		computes the keyed hash using the MD5 hash function
//	HMAC_SHA1  (msg,key)		computes the keyed hash using the SHA1 hash function
//	HMAC_SHA256(msg,key)		computes the keyed hash using the SHA256 hash function
//	HMAC_FP32  (msg,key)		computes the keyed hash using the HRC4.FP32 method
//	HMAC_FP64  (msg,key)		computes the keyed hash using the HRC4.FP64 method
//
//	cypher  (msg,key)		elementary encyphering/decyphering of a single message
//	encypher(msg,key)		encyphering of a single message including its SHA1 hash
//	decypher(msg,key)		inverse of encypher, checks the SHA1 for correct decyphering
//	*cypher2(msg,key1,key2)		same, but with the new init2 initialisation
//
//	encrypt (msg,key,unicode,addmarker) same as encypher, but escapes Unicode in message and converts result to B64
//	decrypt (msg,key,unicode)	inverse of encrypt, including Base64 decoding, SHA1 check, and unescape
//	*crypt2 (msg,key1,key2,...)	same, but with the new init2 initialisation
//
//	betterKey (key)			converts a key to a far better key, repeatedly using the SHA256 algorithm
//
// Instance properties:
//	_s, _i, _j			the PRNG status and its indexes.
//					THESE ARE TOO BE TREATED AS INTERNAL DATA OF THE OBJECT,
//					YOUR CODE SHOULD NEVER DIRECTLY ACCESS THESE PROPERTIES!
//
// Instance methods (prototype methods):
//	init2(key1,key2,reshuffle)	better initialisation of the HRC4 internal status
//	init(key)			(re)initialises the HRC4 Object instance's internal status using the key
//	reshuffle(key)			reshuffles the internal PRNG status using another key w/o re-initing it
//	dump()				returns the internal status as a 258 byte String Object
//	load(string)			loads the internal status where the input is as a 258 byte String Object,
//					as returned by the dump method
//	random(cnt)			returns a number of pseudorandom bytes in a String Object
//	randomReal()			returns a pseudorandom Real Number between 0 (inclusive) and 1 (exclusive)
//	randomString(len)		returns a pseudorandom printable String of Base 64 characters
//
//	cypher  (msg)			elementary en/decyphering of a message
//	encypher(msg)			encyphering of a message including its SHA1 hash
//	decypher(msg)			inverse of encypher, with SHA1 check for correct decyphering
//
//	encrypt (msg,unicode,addmarker)	same as encypher, but escapes Unicode in message and converts result to Base64
//	decrypt (msg,unicode)		inverse of encrypt, including Base64 decoding, SHA1 check, and unescape
//
// Note:
//	The Constructor Properties and Methods are readily available once this 'hrc4.js' file has been included,
//	HRC4 Object Instances must be created by code in the page or script that includes this 'hrc4.js' file.
//
// Usage:
//	It is recommended to use the 'encrypt' and 'decrypt' methods for securing your messages.
//	The 'cypher', 'encypher' and 'decypher' methods can also be used if desired,
//	but it is better to consider them as internal methods of the HRC4 Object.
//
//	Static methods can be used directly, without creating an HRC4 Object instance.
//	Examples:
//		var myMessage = "This message is classified"
//		var myKey     = "This key is topsecret"
//		var myCypher  = HRC4.encrypt(myMessage,myKey)
//		var decrypted = HRC4.decrypt(myCypher ,myKey)
//		if (!HRC4.lastDecypherIsGood) ERROR!!!
//
//	For using the prototype methods, an Object instance must be created first.
//	You can define as many different instances as needed, each one initialised with its own key.
//	Examples:
//		Sender:
//			var key = "Thou shalt commit adultery"		// send to receiver in a TRUSTED way!
//			var cryptObj = new HRC4(key)
//			var encrMsg1 = cryptObj.encrypt("The first encrypted message")
//			var encrMsg2 = cryptObj.encrypt("The second encrypted message")
//		Receiver:
//			var cryptObj = new HRC4(key)			// use same key as sender
//			var message1 = cryptObj.decrypt(encrMsg1)	// must decrypt in same order
//			if (!HRC4.lastDecypherIsGood) ERROR!!!
//			var message2 = cryptObj.decrypt(encrMsg2)	// as encryption took place!
//			if (!HRC4.lastDecypherIsGood) ERROR!!!
//
//	PLEASE NOTE THAT KEYS SHOULD ALWAYS BE EXCHANGED IN A TRUSTED WAY! NOT AS PLAIN TEXT VIA INTERNET!
//
// Update history (since 12-Feb-2006):
//	24+28-Feb-2010
//		modified String.ByteChars and the String.fromByte and String.prototype.byteAt methods
//		(some site did not work with IE8, but it did with IE6, Chrome, Firefox)
//	17-Feb-2010
//		implemented HRC4.betterkey(key)
//	12-Feb-2006
//		implemented HRC4.blocksize for faster handling of large Strings
//	07-Apr-2006
//		fixed typing error in definition of function HRC4 (argument "arcourMode" => "arcfourMode")
//	18/30-Aug-2006
//		- fixed minor typos in comments + added note about invalid chars. in 'fromBase64'
//		- added 'wrap' argument to the 'toBase64' method
//		- option to add a marker in 'encrypt' methods for automatic unicode detection by 'decrypt'
//		- changed the to/fromQuotPr methods to use CR+LF as line separator in result
//		- implemented FP32/64 and HMAC_FP32/64
//	01-Sep-2006
//		- changed part of regexpr in toHex/fromHex from "0-9A-Fa-f" to "a-f0-9" with /i modifier
//		- suppress initialisation if 'new HRC4()' is used w/o arguments, intended for restoring status
//		  from persistent userdata (to be implemented).
//	17-Jan-2007
//		- implemented the 'reshuffle' prototype method
//		- implemented the 'dump' prototype method
//	24-Mar-2008
//		implemented load, toDec, and fromDec methods
//	15-Apr-2008
//		implemented the 'init2', 'new2', '*cypher2', and '*crypt2' methods;
//		fixed bug in the 'toDec' method;
//		implemented a 'new1' method for convenience.
//
// DISCLAIMER:
//	HENK REINTS HAS WRITTEN THIS SOFWARE FOR HIS OWN USE, AS A HOBBY, WITHOUT ANY COMMERCIAL INTENTIONS.
//	IT IS PROVIDED AS IS, WITHOUT ANY GUARANTEE FOR CORRECTNESS OR USABILITY. USAGE IS AT YOUR OWN RISK.
//	IN NO WAY SHALL HENK REINTS BE LIABLE FOR ANY NEGATIVE CONSEQUENCES OF USING THIS SOFTWARE.
//
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


// ================================================================================================
// First we must compensate for the weird behaviour of Windows that converts some characters with
// byte values in range 128..255 to 16-bit Unicode characters, such as the Euro-symbol "". Using
// a translation table we can counteract this unwanted behaviour.
// The HRC4 Object uses the 'fromByte' and 'byteAt' methods defined below instead of the standard
// 'fromCharCode' and 'charCodeAt' methods of the String Object.
// The 'fromByte' method is defined without arguments, but it can be called with any number of
// arguments, each one being a byte value (same behaviour as the built-in String.fromCharCode).
// ------------------------------------------------------------------------------------------------

	String.ByteChars = String.fromCharCode
			(128,129,130,131,132,133,134,135,136,137,138,139,140,141,142,143
			,144,145,146,147,148,149,150,151,152,153,154,155,156,157,158,159
			,160,161,162,163,164,165,166,167,168,169,170,171,172,173,174,175
			,176,177,178,179,180,181,182,183,184,185,186,187,188,189,190,191
			,192,193,194,195,196,197,198,199,200,201,202,203,204,205,206,207
			,208,209,210,211,212,213,214,215,216,217,218,219,220,221,222,223
			,224,225,226,227,228,229,230,231,232,233,234,235,236,237,238,239
			,240,241,242,243,244,245,246,247,248,249,250,251,252,253,254,255
			)

	String.fromByte = function ()
	{	for (var x = "", i = 0; i < arguments.length; x += String.fromCharCode(arguments[i++]&255) );
		return x
	}
	String.prototype.byteAt = function (i)
	{	var c = String.ByteChars.indexOf(this.charAt(i))
		return (c < 0 ? this.charCodeAt(i) : c+128) & 255	// added "&255" 2010-02-24
	}


// ################################################################################################
//	N E X T   S E C T I O N   ==   C O N S T R U C T O R   D E F I N I T I O N
// ################################################################################################

// ================================================================================================
// Constructor for the HRC4 object.
// Arguments:
//	key		String with key for initialisation of PRNG status vector,
//			MUST NOT be a zero length String.
//	arcfourMode	optional, Boolean indicating if original arcfour mode status
//			initialisation should be done instead of the improved HRC4 mode.
//			if omitted it defaults to False, i.e. HRC4 mode is used by default.
// Returns:
//	<nothing>
// Usage:
//	var cryptObj = new HRC4 (myVerySecretKey)
// Note:
//	properties and methods with a name starting with "_" are to be considered INTERNAL,
//	not to be directly manipulated or invoked by the "outside world".
// ------------------------------------------------------------------------------------------------
	function HRC4 (key, arcfourMode)
	{	this._i = NaN, this._j = NaN, this._s = []
		if	(arguments.length > 1)	this.init(key,arcfourMode)
		else if	(arguments.length > 0)	this.init(key)
	}


// ################################################################################################
//	N E X T   S E C T I O N   ==   C O N S T R U C T O R   P R O P E R T I E S
// ################################################################################################

// ================================================================================================
// The 'lastDecypherIsGood' property is a Boolean that is set by all 'decypher' and 'decrypt'
// methods, both static and prototype (well, as a matter of fact, it is set only by the
// 'prototype.decypher' method, which is used by the others).
// It is set to True if the SHA1 check passes, False if it fails.
// After every decryption the 'HRC4.lastDecypherIsGood' property can (and should) be checked.
// Example (assuming you have a web page containing confident information):
//   HTML:
//	<input type=password name="thePassword">
//	<input type=submit onclick="return checkPassword(window.event)">
//   javascript:
//	function checkPassword(theEvent)
//	{	// assume encrypted data are available in GLOBAL variable 'encryptedData'
//		// next will set GLOBAL variable 'decryptedData' to contain the result.
//		decryptedData = HRC4.decrypt(encryptedData,document.all.thePassword.value)
//		if (!HRC4.lastDecypherIsGood)
//		{	alert ("Incorrect password!")	// message if wrong
//			theEvent.cancelBubble = true	// stop all further button click handling
//		}
//		return HRC4.lastDecypherIsGood		// pass status back to the 'Submit' button
//	}
// Note:
//	although javascript allows you to modify it directly,
//	this property should be treated as READ-ONLY.
// ------------------------------------------------------------------------------------------------
	HRC4.lastDecypherIsGood = false

// ================================================================================================
// The '_MD5table' property is to be considered an INTERNAL HRC4 property, do not modify it!
// It contains a list of numbers needed for computing the MD5 hash of a message.
// It is the "T" array needed for MD5 computation as defined in RFC1321, computed as follows:
//	for (var T=[],i=0; i<64; T[i++]=Math.floor(Math.abs(Math.sin(i))*0x100000000) );
// ------------------------------------------------------------------------------------------------
	HRC4._MD5table =
	[0xd76aa478,0xe8c7b756,0x242070db,0xc1bdceee,0xf57c0faf,0x4787c62a,0xa8304613,0xfd469501
	,0x698098d8,0x8b44f7af,0xffff5bb1,0x895cd7be,0x6b901122,0xfd987193,0xa679438e,0x49b40821
	,0xf61e2562,0xc040b340,0x265e5a51,0xe9b6c7aa,0xd62f105d,0x02441453,0xd8a1e681,0xe7d3fbc8
	,0x21e1cde6,0xc33707d6,0xf4d50d87,0x455a14ed,0xa9e3e905,0xfcefa3f8,0x676f02d9,0x8d2a4c8a
	,0xfffa3942,0x8771f681,0x6d9d6122,0xfde5380c,0xa4beea44,0x4bdecfa9,0xf6bb4b60,0xbebfbc70
	,0x289b7ec6,0xeaa127fa,0xd4ef3085,0x04881d05,0xd9d4d039,0xe6db99e5,0x1fa27cf8,0xc4ac5665
	,0xf4292244,0x432aff97,0xab9423a7,0xfc93a039,0x655b59c3,0x8f0ccc92,0xffeff47d,0x85845dd1
	,0x6fa87e4f,0xfe2ce6e0,0xa3014314,0x4e0811a1,0xf7537e82,0xbd3af235,0x2ad7d2bb,0xeb86d391
	]

// ================================================================================================
// The 'blocksize' property is used when long strings are handled. Building a string character by
// character is typically done using the 'x += char' operation, but as far as I have found out,
// this does not really APPEND char to x but it copies both x and char into a new String which is
// then assigned to x. This means it gets slower and slower as x is growing. Therefore the building
// of result strings is done blockwise. Each substring will be stored in an intermediate Array
// (which is also inefficient, so it too should not grow too large) which is finally joined to the
// complete result. The value below determines the blocksize, which you can change as desired.
// A larger blocksize makes the String building slower, but the Array faster, and vice versa.
// ------------------------------------------------------------------------------------------------
	HRC4.blocksize = 1024


// ################################################################################################
//	N E X T   S E C T I O N   ==   P R O T O T Y P E   M E T H O D S
// ################################################################################################

// ================================================================================================
// New Prototype method (14-APR-2008) for advanced initialisation of the internal PRNG status
// (PRNG = Pseudo Random Number Generator) making it hopefully far less crackable.
// It uses 2 keys and does MUCH more initialisation rounds for better inpredictability of the
// initial status (which in the original RC4 is always 0..255, to my opinion part of the cause of
// the Fluhrer, Mantin and Shamir attack). Hopefully the initial keys are hidden much better than
// the in original RC4. This method should be called separately, i.e.:
//
//		var myHRC4 = new HRC4()	// create empty (uninitialised) HRC4 Object instance
//		myHRC4.init2(key1,key2)	// init it using the new initialisation method
//
//		(or better/easier: use the 'new2' method which does this for your)
// Arguments:
//	key1,key2	Strings with the 2 keys for initialisation of PRNG status vector,
//			MUST NOT be zero length Strings, the longer the better,
//			MUST have completely different content
//			and preferably have different length.
//	reshuffle	optional argument,
//			if evaluated as true then the primary status initialisation to 0..255
//			is skipped, giving this method the same sort of behaviour as the
//			old-style 'reshuffle' method.
// Returns:
//	<nothing>
// ------------------------------------------------------------------------------------------------
	HRC4.prototype.init2 = function (key1, key2, reshuffle)
	{	var f,k,n,t, s = this._s, i = this._i, j = this._j
		var K1 = HRC4.SHA256(key1) + key1, L1 = K1.length	// include hash of first key
		var K2 = HRC4.SHA256(key2) + key2, L2 = K2.length	// and of second
		var L  = 2*Math.max(L1+L2,512)				// set a minimum no. of KSA rounds
		f = (reshuffle ? false : true)					// need first init from 0..255?
		if (!isFinite(i) || i < 0 || i > 255 || i%1 != 0) f = true	// status invalid -> must init
		if (!isFinite(j) || j < 0 || j > 255 || j%1 != 0) f = true	// status invalid -> must init
		if (f) for (i=0; i<256; s[i]=i++);				// init status if necs or wanted
		for (n = 0; n<L; n++)					// simple initial mixup using both keys
		{	i = s[K1.byteAt(n%L1)&255]
			j = s[K2.byteAt(n%L2)&255]
			t = s[i], s[i] = s[j], s[j] = t
		}
		for (n = 0; n < L; n++)					// initial freewheeling
		{	i = (i + 1)&255, j = (j + s[i] ) & 255
			t = s[i], s[i] = s[j], s[j] = t
		}
		for (k = 0, n = 0; n < L; i++, k++, n++)		// use key1 in more or less normal way
		{	i &= 255, j = (j + s[i] + K1.byteAt(k%L1)&255)&255
			t = s[i], s[i] = s[j], s[j] = t
		}
		for (n = 0; n < L; n++)					// intermediate freewheeling
		{	i = (i + 1)&255, j = (j + s[i] ) & 255
			t = s[i], s[i] = s[j], s[j] = t
		}
		for (k = 0, n = 0; n < L; i++,k++,n++)			// use key2 like key1
		{	i &= 255, j = (j + s[i] + K2.byteAt(k%L2)&255)&255
			t = s[i], s[i] = s[j], s[j] = t
		}
		for (n = 0; n < L; n++)					// final freewheeling
		{	i = (i + 1)&255, j = (j + s[i] ) & 255
			t = s[i], s[i] = s[j], s[j] = t
		}
		this._i = i, this._j = j
	}

// ================================================================================================
// Prototype method to initialise the internal status vector of the PRNG.
// (PRNG = Pseudo Random Number Generator).
// Called automatically by the constructor of the HRC Object,
// but can be used to re-initialise an HRC4 Object's internal PRNG status.
// Arguments:
//	key		String with key for initialisation of PRNG status vector,
//			MUST NOT be a zero length String.
//	arcfourMode	optional, Boolean indicating if original arcfour mode status
//			initialisation should be done instead of the improved HRC4 mode.
//			if omitted it defaults to False, i.e. HRC4 mode is used by default.
// Returns:
//	<nothing>
// ------------------------------------------------------------------------------------------------
	HRC4.prototype.init = function (key, arcfourMode)
	{	var i,j,k,n,t, s = this._s, K = key.length
		var M = (arcfourMode ? 256 : Math.max (512, 2*K) )	// no. of KSA rounds
		var N = (arcfourMode ?   0 : 997)			// no. of freewheel rounds
		for (i = 0; i < 256; s[i] = i++);
		for (i = 0, j = 0, k = 0, n = 0; n < M; i++, k++, n++)
		{	i &= 255, j = (j + s[i] + key.byteAt(k%K)&255) & 255
			t = s[i], s[i] = s[j], s[j] = t
		}
		for (i = 0, j = 0, n = 0; n < N; n++)
		{	i = (i + 1) & 255, j = (j + s[i] ) & 255
			t = s[i], s[i] = s[j], s[j] = t
		}
		this._i = i, this._j = j
	}

// ================================================================================================
// Prototype method to reshuffle the internal status vector based on a new key,
// without first reinitialising the vector. Intended use is to modify the status based on
// unpredictable events, thus producing a less predictable (or even really unpredictable and
// irreproducible) random number generator.
// Arguments:
//	key		String with key for reshuffling the PRNG status vector
// Returns:
//	<nothing>
// ------------------------------------------------------------------------------------------------
	HRC4.prototype.reshuffle = function (key)
	{	var i = this._i, j = this._j, s = this._s, k = 0, K = key.length, t
		for (; k < K; i&=255, j=(j+s[i]+key.byteAt(k)&255)&255, t=s[i], s[i]=s[j], s[j]=t, i++, k++);
		this._i = i, this._j = j
	}

// ================================================================================================
// Prototype method to return the internal PRNG status vector as a String Object
// containing BINARY data. Should be converted to hex or b64 before printing.
// Arguments:
//	<none>
// Returns:
//	258 character String with byte values of all elements of s, then i and j
// 24-Mar-2008: also implemented 'load' method, expecting result of dump as input,
// it will reload the internal status with the result of dump.
// Arguments:
//	258 character String object with byte values as returned by the dump method.
// Returns:
//	nothing.
// ------------------------------------------------------------------------------------------------
	HRC4.prototype.dump = function()
	{	var s = this._s, k = 0, K = s.length, result = ""
		for (; k < K; result += String.fromByte(s[k++]) );
		result += String.fromByte(this._i)
		result += String.fromByte(this._j)
		return result
	}
	HRC4.prototype.load = function (x)
	{	for (var k = 0; k < 256; this._s[k] = x.byteAt(k), k++);
		this._i = x.byteAt(k++), this._j = x.byteAt(k)
	}

// ================================================================================================
// Prototype method to obtain a (list of) pseudorandom byte value(s).
// This is not for encryption, just a "free bonus" method that can be used to generate a
// reproducible series of pseudorandom byte values based on a arbitrary length initialisation key.
// Each successive call will continue the series at the point where the last call left.
// Arguments:
//	cnt	optional, Number, no. of bytes to be returned
//		if omitted it defaults to 1
// Returns:
//	String with the requested number of bytes
// Notes:
//	o As far as I can find out on the Internet, the arcfour Pseudo Random Number Generator
//	  is a fairly good one, but not perfect for performing high level statistics.
//	o The 'randomString' method also returns a String Object, but that String contains Base 64
//	  characters only so it is a printable text, whilst 'random' returns a String that may
//	  contain any byte value from 0 to 255, which is not to be considered as a printable text.
// ------------------------------------------------------------------------------------------------
	HRC4.prototype.random = function (cnt)
	{	var i = this._i, j = this._j, s = this._s, n
		var M = (arguments.length > 0 ? cnt : 1)
		var result = [], temp = "", m = 0
		for (n = 0; n < M; n++)
		{	i = (i + 1) & 255, j = (j + s[i] ) & 255
			t = s[i], s[i] = s[j], s[j] = t
			temp += String.fromByte(s[(s[i]+s[j])&255])
			if (temp.length >= HRC4.blocksize) {result[m++] = temp, temp = ""}
		}
		this._i = i, this._j = j
		if (temp.length > 0) result[m] = temp
		return result.join("")
	}

// ================================================================================================
// Prototype method to obtain a pseudorandom Real Number between 0 (inclusive) and 1 (exclusive).
// In fact just like the built-in 'Math.random()' method, but this one is based on the HRC4 PRNG.
// It is not needed for encryption, just another "free bonus" method for generating a reproducible
// series of pseudorandom Real Numbers seeded with an arbitrary length initialisation key.
// Each successive call will continue the series at the point where the last call left.
// Arguments:
//	<none>
// Returns:
//	A pseudorandom Real Number between 0 (inclusive) and 1 (exclusive)
// Note:
//	The Number has a 53-bit precision, which is the highest achievable in JavaScript
// Technical detail:
//	These 53 bits are produced by generating 7 random bytes of which 8 or 7 bits are taken,
//	which is achieved by toggling the mask in variable m between 0xff and 0x7f. This toggling
//	is done by iteratively XOR-ing m with 0x80. In this way the numbers of bits taken are:
//	8,7,8,7,8,7,8 which adds up to a total of 53.
// ------------------------------------------------------------------------------------------------
	HRC4.prototype.randomReal = function ()
	{	var i = this._i, j = this._j, s = this._s, n, x = 0, m = 0xff
		for (n = 0; n < 7; n++)
		{	i = (i + 1) & 255, j = (j + s[i] ) & 255
			t = s[i], s[i] = s[j], s[j] = t
			x = x*(m+1) + (s[(s[i]+s[j])&255]&m), m ^= 0x80
		}
		this._i = i, this._j = j
		return x/0x20000000000000
	}

// ================================================================================================
// Prototype method to obtain a String with pseudorandom printable text of a given length.
// All characters are from the Base 64 character set.
// This is not for encryption, just another "free bonus" method that can be used to generate a
// reproducible series of pseudorandom Strings based on a arbitrary length initialisation key.
// Each successive call will continue the series at the point where the last call left.
// Arguments:
//	len	Number, length that the resulting String should have
// Returns:
//	String with the requested number of Base 64 characters
// Note:
//	The 'random' method also returns a String Object, but that String may contain any byte
//	value from 0 to 255, not to be considered as printable text, whilst 'randomString' returns
//	a String of Base 64 characters only, which is printable.
// ------------------------------------------------------------------------------------------------
	HRC4.prototype.randomString = function (len)
	{	var D = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/"
		var i = this._i, j = this._j, s = this._s, n
		var result = [], temp = "", m = 0
		for (n = 0; n < len; n++)
		{	i = (i + 1) & 255, j = (j + s[i] ) & 255
			t = s[i], s[i] = s[j], s[j] = t
			temp += D.charAt(s[(s[i]+s[j])&255]&0x3f)
			if (temp.length >= HRC4.blocksize) {result[m++] = temp, temp = ""}
		}
		this._i = i, this._j = j
		if (temp.length > 0) result[m] = temp
		return result.join("")
	}

// ================================================================================================
// Prototype methods to encypher or decypher a message using the current PRNG status.
// 'cypher' is the core cyphering algorithm (ARCFOUR).
// Arguments:
//	msg		String with message to be encyphered or cypher to be decyphered.
//	unicode		optional (**crypt only), Boolean whether to use unicode/byte conversion
//			instead of escape/unescape, which reduces the size of the cypher if msg
//			contains LOTS of Unicode characters above 255.
//			if False or omitted then escape/unescape is used.
//	addmarker	optional, for 'encrypt' only and only used if the 'unicode' arg. is given
//			and true: if 'addmarker' is evaluated as true then result is prefixed with
//			a marker ("$\") to identify it as being converted using the unicode/byte
//			conversion, so 'decrypt' will be able to automatically detect it, i.e. you
//			do not need to specify its 'unicode' argument. It has been implemented this
//			way for backward compatibility with older versions of the HRC4 Object.
// Returns:
//	o String with encyphered or decyphered result.
//	o The 'decypher' and 'decrypt' methods also set the HRC4.lastDecypherIsGood property
//	  to either True or False, indicating whether the decyphering key was correct or not.
// Notes:
//	o the 'encypher' method includes the SHA1 hash of the message in the encyphered result,
//	  the 'decypher' method checks for correct decyphering using this SHA1 hash.
//	o the '**crypt' methods do the same as the '**cypher' methods,
//	  but they do implicit conversion of Unicode characters above 255
//	  and they convert the cypher to/from Base64 notation.
// ------------------------------------------------------------------------------------------------
	HRC4.prototype.cypher = function (msg)
	{	var i = this._i, j = this._j, s = this._s, M = msg.length, c, k, n
		var result = [], temp = "", m = 0
		for (n = 0; n < M; n++)
		{	i = (i + 1) & 255, j = (j + s[i] ) & 255
			t = s[i], s[i] = s[j], s[j] = t
			c = msg.byteAt(n) & 255
			k = (s[i] + s[j] ) & 255
			temp += String.fromByte (c ^ s[k] )
			if (temp.length >= HRC4.blocksize) {result[m++] = temp, temp = ""}
		}
		this._i = i, this._j = j
		if (temp.length > 0) result[m] = temp
		return result.join("")
	}
	HRC4.prototype.encypher = function (msg)
	{	return this.cypher (HRC4.SHA1 (msg) + msg)
	}
	HRC4.prototype.decypher = function (msg)
	{	var tmp = this.cypher (msg), txt = tmp.slice(20)
		HRC4.lastDecypherIsGood = (HRC4.SHA1 (txt) == tmp.slice(0,20) )
		return txt
	}
	HRC4.prototype.encrypt = function (msg,unicode,addmarker)
	{	if (unicode && addmarker)	return	"$\\" + HRC4.toBase64(this.encypher(HRC4.unicodeToBytes(msg)))
		else if (unicode)		return		HRC4.toBase64(this.encypher(HRC4.unicodeToBytes(msg)))
		else				return		HRC4.toBase64(this.encypher(HRC4.escape        (msg)))
	}
	HRC4.prototype.decrypt = function (msg,unicode)
	{	if (msg.slice(0,2) == "$\\" || unicode)
			return	HRC4.bytesToUnicode(this.decypher(HRC4.fromBase64(msg)))
		else	return	unescape           (this.decypher(HRC4.fromBase64(msg)))
	}


// ################################################################################################
//	N E X T   S E C T I O N   ==   S T A T I C   M E T H O D S
// ################################################################################################

// ================================================================================================
// New Static method (16-APR-2008) for easily obtaining a new HRC4 Object instance
// initialised using the OLD 'init' method.
// Implemented just for analogy to the 'new2' method below.
// Effectively it's the same as the usual "new HRC4(...)" syntax.
// ------------------------------------------------------------------------------------------------
	HRC4.new1 = function (key, arcfourMode)
	{	var x = new HRC4()
		if	(arguments.length > 1)	x.init(key,arcfourMode)
		else if	(arguments.length > 0)	x.init(key)
		return x
	}

// ================================================================================================
// New Static method (14-APR-2008) for easily obtaining a new HRC4 Object instance
// initialised using the NEW 'init2' method with 2 keys
// Usage:
//	instead of:	var myHRC4 = new HRC4(); myHRC4.init2(key1,key2)
//	simply use:	var myHRC4 = HRC4.new2(key1,key2)
// ------------------------------------------------------------------------------------------------
	HRC4.new2 = function (key1, key2)
	{	var x = new HRC4()
		x.init2(key1, key2)
		return x
	}

// ================================================================================================
// STATIC method for escaping Unicode characters in a String.
// It is a "soft" variant of JavaScript's built-in 'escape' method.
// Only the percent sign and character codes greater than 255 are escaped.
// Arguments:
//	txt	String to be escaped.
// Returns:
//	String with escaped result.
// ------------------------------------------------------------------------------------------------
	HRC4.escape = function (txt)
	{	var i, c, cc, result = [], temp = "", m = 0
		for (i = 0; i < txt.length; i++)
		{	c = txt.charAt(i), cc = c.byteAt(0)
			temp += (c == "%" || cc > 255 ? escape(c) : c)
			if (temp.length >= HRC4.blocksize) {result[m++] = temp, temp = ""}
		}
		if (temp.length > 0) result[m] = temp
		return result.join("")
	}

// ================================================================================================
// STATIC methods for encoding a String of byte values in hexadecimal notation and vice versa.
// Arguments:
//	msg	String with message to be encoded or decoded.
//	sep	optional, String to be used as separator between individual hex byte values,
//		any hexadecimal digits are removed from this separator before use.
//		if omitted all hexadecimal digits will be concatenated without separation.
// Returns:
//	String with encoded or decoded result.
// Note:
//	The 'fromHex' method allows a reasonably free format. Every single comma or non-hex
//	pattern is interpreted as a separator of hexadecimal substrings. If such a substring
//	is blank (i.e. nothing between 2 consecutive commas) it is interpreted as a single zero
//	byte (so ",,," is converted to 4 consecutive null bytes). If a substring has odd length a
//	"0" will be PREpended to it, mainly to allow bytes to be represented by single hex digits
//	separated by non-hex characters (longer substrings SHOULD have an even length).
// Update 24-Mar-2008:
//	the toDec and fromDec methods are equivalent to toHex and fromHex
//	keep in mind that input to to fromDec should have some separator!
// ------------------------------------------------------------------------------------------------
	HRC4.toHex = function (msg, sep)
	{	var S=(arguments.length>1?(""+sep).replace(/[a-f0-9]+/gi,""):"")
		var i=0, L=msg.length, s="", result = [], temp = "", m = 0
		for (; i<L; s=S)
		{	temp+=s+(256|(msg.byteAt(i++)&255)).toString(16).slice(-2)
			if (temp.length >= HRC4.blocksize) {result[m++] = temp, temp = ""}
		}
		if (temp.length > 0) result[m] = temp
		return result.join("")
	}
	HRC4.fromHex = function (msg)
	{	var i,j,L,H,A=msg.replace(/[^a-f0-9,]+/gi,",").split(",")
		var result = [], temp = "", m = 0
		for (i=0; i<A.length; i++)
		{	L=A[i].length, H=(L<1?"00":L%2==1?"0"+A[i]:A[i])
			for (j=0; j<H.length; )
			{	temp+=String.fromByte(parseInt(H.slice(j++,++j),16))
				if (temp.length >= HRC4.blocksize) {result[m++] = temp, temp = ""}
			}
		}
		if (temp.length > 0) result[m] = temp
		return result.join("")
	}
	HRC4.toDec = function (msg, sep)
	{	var S=(arguments.length>1?(""+sep).replace(/[0-9]+/gi,""):"")
		var i=0, L=msg.length, s="", result = [], temp = "", m = 0
		if (S == "") S = ",";
		for (; i<L; s=S)
		{	temp+=s+(msg.byteAt(i++)&255).toString(10)
			if (temp.length >= HRC4.blocksize) {result[m++] = temp, temp = ""}
		}
		if (temp.length > 0) result[m] = temp
		return result.join("")
	}
	HRC4.fromDec = function (msg)
	{	var i,j,L,H,A=msg.replace(/[^0-9,]+/gi,",").split(",")
		var result = [], temp = "", m = 0
		for (i=0; i<A.length; i++)
		{	if (A[i].length < 1) A[i] = "0";
			temp+=String.fromByte(parseInt(A[i],10)&255)
			if (temp.length >= HRC4.blocksize) {result[m++] = temp, temp = ""}
		}
		if (temp.length > 0) result[m] = temp
		return result.join("")
	}


// ================================================================================================
// STATIC methods for encoding a String of byte values in Base64 notation and vice versa.
// Arguments:
//	msg	String with message to be encoded or decoded.
//	wrap	optional argument,
//		if boolean true then the result will be wrapped at a width of 76, which is the default
//		wrap width according to the MIME specifications, if boolean false then it will not be
//		wrapped (which is default if this arg. is omitted), else the wrap width will be set to
//		1*wrap (which might result in NaN if it is not a number, causing no wrapping)
//	nopad	optional argument,
//		if evaluated as true then the result will not be padded with equals signs (the MIME
//		specification prescribes padding with equals signs until the result length equals a
//		multiple of 4 characters).
// Returns:
//	String with Base64 encoded or decoded result.
//	if wrapped, the line separator will be "\r\n", the good old carriage return + linefeed.
// Notes:
//	o according to the MIME specification, Base64 encoded data should be split into lines of 76
//	  characters, with CR+LF as line separator.
//	o the 'fromBase64' method ignores any invalid characters in the input string.
//	o if the 'nopad' argument is given, then 'wrap' should also be specified as true, false, or a number.
// Reference:
//	RFC2045: Multipurpose Internet Mail Extensions (MIME) Part One: Format of Internet Message Bodies
//	http://www.rfc-editor.org/rfc/rfc2045.txt
// ------------------------------------------------------------------------------------------------
	HRC4.toBase64 = function (msg,wrap,nopad)
	{	var D = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/"
		var i,j,k,m,n,K, L = msg.length
		var result = [], temp = "", mm = 0
		var W = (arguments.length < 2 ? NaN : wrap === true ? 76 : wrap === false ? NaN : 1*wrap)
		for (i=0,n=0; i<L; )
		{	for (j=0,k=0,K=0; j<3 && i<L; K<<=8,K|=(msg.byteAt(i++)&255),k+=6,j++);
			for (m=j; m<3; K<<=8,k+=6,m++);
			for (m=18-6*j; k>=m; k-=6)
			{	n++
				temp += D.charAt((K>>>k)&63) + (n%W == 0 ? "\r\n" : "")
				if (temp.length >= HRC4.blocksize) {result[mm++] = temp, temp = ""}
			}
		}
		if (nopad) j = 3;
		for (; j<3; temp+="=",j++);
		if (temp.length > 0) result[mm] = temp
		return result.join("")
	}
	HRC4.fromBase64 = function (msg)
	{	var D = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/"
		var c,i,j,k,m,K, L = msg.length
		var result = [], temp = "", mm = 0
		for (i=0; i<L; )
		{	for (j=0,K=0; j<4 && i<L;) {c=D.indexOf(msg.charAt(i++)); if (c>=0) {K<<=6,K|=(c),j++} }
			for (m=j; m<4; K<<=6,m++);
			for (k=16; j>1; k-=8,j--)
			{	temp+=String.fromByte((K>>>k)&255)
				if (temp.length >= HRC4.blocksize) {result[mm++] = temp, temp = ""}
			}
		}
		if (temp.length > 0) result[mm] = temp
		return result.join("")
	}

// ================================================================================================
// STATIC methods for encoding and decoding a String in Quoted Printable notation.
// Arguments:
//	msg		String with message to be encoded or decoded.
//	multiline	(toQuotPr only), optional Boolean to indicate msg is multiline.
// Returns:
//	String with encoded or decoded result.
// Notes:
//	o The 'multiline' argument of the 'toQuotPr' method causes the following:
//	  - each CR+LF pair is treated as a line separator,
//	    as well as every single CR or LF (after converting every CR+LF to LF);
//	  - the message is split into separate lines at the line separators;
//	  - each line is encoded;
//	  - the result is concatenated to a single result with "\r\n" (CRLF) as line separotor.
//	  If multiline is False or omitted then all CR's and LF's in the message are encoded
//	  as '=0D' and '=0A' respectively.
//	o In Quoted Printable, an 'equals' sign at the very end of a line indicates a 'soft'
//	  linebreak, i.e. the decoded result of the next line should be appended to it without
//	  an intermediate line separator. If the input message of fromQuotPr does not end with a
//	  trailing 'equals' sign, then fromQuotPr appends a CRLF to the result, indicating a hard
//	  end of line. This means that a decoded result should be output using a 'Write' method
//	  instead of a 'WriteLine' method.
//	o Both methods use the CRLF character pair as linebreak in the result, as required by the
//	  MIME specification.
//	o Quoted Printable is intended to encode plain text containing SOME non-printable and/or
//	  special non-7-bit characters. For encoding true binary data, Base64 is recommended.
// Reference:
//	RFC2045: Multipurpose Internet Mail Extensions (MIME) Part One: Format of Internet Message Bodies
//	http://www.rfc-editor.org/rfc/rfc2045.txt
// ------------------------------------------------------------------------------------------------
	HRC4.toQuotPr = function (msg, multiline)
	{	if (multiline)
		{	var tmp = msg.replace(/\r\n/g,"\n").replace(/\r/g,"\n").split("\n")
			for (var i = 0; i < tmp.length; tmp[i]=HRC4.toQuotPr(tmp[i]), i++);
			return tmp.join("\r\n")
		}
		var c, i=0, W=76, M=msg.length, L=0, softline="=\r\n"
		var result = [], temp = "", m = 0
		for (; i < M; )
		{	c = msg.charAt(i++); if (i==M) W++
			if (c < " " || c == "=" || c > "~" || (c == " " && i == M) ) c="="+HRC4.toHex(c).toUpperCase()
			L+=c.length; if (L<W) {temp+=c} else {temp+=softline+c, L=c.length}
			if (temp.length >= HRC4.blocksize) {result[m++] = temp, temp = ""}
		}
		if (temp.length > 0) result[m] = temp
		return result.join("")
	}
	HRC4.fromQuotPr = function (msg)
	{	var  c, i=0, y=msg.replace(/\r\n/g,"\n").replace(/\r/g,"\n").replace(/=\n/g,"")
		var result = [], temp = "", m = 0
		var soft = y.search(/=$/); if (soft>=0) y=y.replace(/=$/,"")
		for (; i<y.length; )
		{	c=y.charAt(i++)
			temp+=(c!="="?c:HRC4.fromHex(y.slice(i++,++i)))
			if (temp.length >= HRC4.blocksize) {result[m++] = temp, temp = ""}
		}
		if (soft<0) temp += "\r\n"
		if (temp.length > 0) result[m] = temp
		return result.join("")
	}

// ================================================================================================
// STATIC methods for converting a Unicode String to separate bytes and vice versa.
// Arguments:
//	msg	String with message to be converted.
// Returns:
//	String with converted result.
// Note:
//	These methods are useful if a message contains LOTS of Unicode characters with charCodes
//	above 255. The 'escape' method would expand every character to 6 bytes ("%uxxxx" where
//	xxxx is the hexadecimal representation of the character code), 'unicodeToBytes' simply
//	splits each character code into 2 bytes.
// ------------------------------------------------------------------------------------------------
	HRC4.unicodeToBytes = function (msg)
	{	var c,i, result = [], temp = "", m = 0
		for (i=0; i<msg.length;)
		{	c=msg.charCodeAt(i++)
			temp+=String.fromByte(c&255,c>>>8)
			if (temp.length >= HRC4.blocksize) {result[m++] = temp, temp = ""}
		}
		if (temp.length > 0) result[m] = temp
		return result.join("")
	}
	HRC4.bytesToUnicode = function (msg)
	{	var i, result = [], temp = "", m = 0
		for (i=0; i<msg.length;)
		{	temp+=String.fromCharCode((msg.byteAt(i++)&255)|((msg.byteAt(i++)&255)<<8))
			if (temp.length >= HRC4.blocksize) {result[m++] = temp, temp = ""}
		}
		if (temp.length > 0) result[m] = temp
		return result.join("")
	}

// ================================================================================================
// STATIC method to compute the MD5 hash of a message.
// Arguments:
//	msg	String with message to be hashed.
// Returns:
//	String containing the raw binary hash value.
// Notes:
//	o The result is a String containing the raw BINARY bytes of the hash as character codes.
//	  It should be considered as non-printable data. Common use in Internet applications is to
//	  convert the value to hexadecimal or Base64 representation, but JavaScript's built-in
//	  'escape' method can also be used to display the hash value in readable form.
//	o The sample c-program of the MD5 algorithm in rfc1321 is SEVENTEEN PAGES,
//	  now take look at the code below...
//	o The code below is indeed quite compact, but a speed gain by a factor 2 or 3 might be
//	  achieved if all functions are recoded in-line. But since MD5 is becoming more and more
//	  obsolete because it is no longer considered really secure (collisions are known), I will
//	  not put much effort in such recoding. For hashing short messages the code below is
//	  perfect, for hashing large files you should use some .exe program for computing the hash.
//	  Robin Keir has a good one, see http://www.keir.net/md5file.html
// Reference:
//	RFC1321: The MD5 Message-Digest Algorithm
//	http://www.rfc-editor.org/rfc/rfc1321.txt
// ------------------------------------------------------------------------------------------------
	HRC4.MD5 = function (msg)
	{	function LongWordToString(val)
		{ for(var r="",v=(val<0?val+0x100000000:val),i=0;i<4;r+=String.fromByte(v&255),v>>>=8,i++);
		  return r
		}
		function stringToLongWords(msg)
		{ for(var r=[],i=0,k=0;i<msg.length;)
		  {for(var v=0,j=0;j<32;v|=(msg.byteAt(i++)&255)<<j,j+=8);r[k++]=v}
		  return r
		}
		var m = 0x100000000, L=msg.length, L1=L%64, L0=L-L1, L2=64-L1, len=L*8, llo=len%m, lhi=(len-llo)/m
		var b128 = String.fromByte(128), nullbyte = String.fromByte(0)
		var pad = (b128+(new Array(64)).join(nullbyte)).slice(0,1+((959-(len&511))&511)>>>3)
		pad += LongWordToString(llo) + LongWordToString(lhi)
		var N = (L+pad.length)/4, M=0xffffffff, A=0x67452301, B=0xefcdab89, C=0x98badcfe, D=0x10325476
		var Q = [7,12,17,22,5,9,14,20,4,11,16,23,6,10,15,21], T = HRC4._MD5table
		function blk(k,x)  {return k<L0 ? x.substr(k,64) : k<L ? x.slice(k)+pad.slice(0,L2) : pad.slice(k-L) }
		function F (x,y,z) {P = (P+1)&15; return (x & y) | (~x & z) }
		function G (x,y,z) {P = (P+5)&15; return (x & z) | (y & ~z) }
		function H (x,y,z) {P = (P+3)&15; return x ^ y ^ z}
		function I (x,y,z) {P = (P+7)&15; return y ^ (x | ~z) }
		function R (w,n)   {var W = w & M; return ( (W << n) | (W >>> (32 - n) ) ) & M}
		function S (f,a,b,c,d) {return (b + R (a + f(b,c,d) + W[P] + T[J], Q[((J>>>4)<<2)|(J++)&3] ) )&M}
		for (var k = 0; k < N; k += 16)
		{	var AA = A, BB = B, CC = C, DD = D, J = 0, W = stringToLongWords(blk(k<<2,msg))
			for (var P=15,i=0; i<4; A=S(F,A,B,C,D), D=S(F,D,A,B,C), C=S(F,C,D,A,B), B=S(F,B,C,D,A), i++);
			for (var P=12,i=0; i<4; A=S(G,A,B,C,D), D=S(G,D,A,B,C), C=S(G,C,D,A,B), B=S(G,B,C,D,A), i++);
			for (var P= 2,i=0; i<4; A=S(H,A,B,C,D), D=S(H,D,A,B,C), C=S(H,C,D,A,B), B=S(H,B,C,D,A), i++);
			for (var P= 9,i=0; i<4; A=S(I,A,B,C,D), D=S(I,D,A,B,C), C=S(I,C,D,A,B), B=S(I,B,C,D,A), i++);
			A = (A + AA) & M; B = (B + BB) & M; C = (C + CC) & M; D = (D + DD) & M;
		}
		return LongWordToString(A)+LongWordToString(B)+LongWordToString(C)+LongWordToString(D)
	}

// ================================================================================================
// STATIC methods to compute the SHA1 or SHA256 hash of a message.
// Arguments:
//	msg	String with message to be hashed.
// Returns:
//	String containing the raw binary hash value.
// Note:
//	The result is a String containing the raw BINARY bytes of the hash as character codes.
//	It should be considered as non-printable data. Common use in Internet applications is to
//	convert the value to hexadecimal or Base64 representation, but JavaScript's built-in
//	'escape' method can also be used to display the hash value in readable form.
// References:
//	o FIPS PUB 180-2: Secure Hash Standard (SHS)
//	  http://csrc.nist.gov/publications/fips/fips180-2/fips180-2withchangenotice.pdf
//	o RFC3174: US Secure Hash Algorithm 1 (SHA1)
//	  http://www.rfc-editor.org/rfc/rfc3174.txt
// ------------------------------------------------------------------------------------------------
	HRC4.SHA1 = function (msg)
	{	function bigEndianLongWordToString(val)
		{ for(var r="",v=(val<0?val+0x100000000:val),i=0;i<4;r=String.fromByte(v&255)+r,v>>>=8,i++);
		  return r
		}
		function stringToBigEndianLongWords(msg)
		{ for(var r=[],i=0,k=0;i<msg.length;)
		  {for(var v=0,j=0;j<4;v<<=8,v|=(msg.byteAt(i++)&255),j++);r[k++]=v}
		  return r
		}
		var b128 = String.fromByte(128), nullbyte = String.fromByte(0)
		var m = 0x100000000, L=msg.length, L1=L%64, L0=L-L1, L2=64-L1, len=L*8, llo=len%m, lhi=(len-llo)/m
		var p = (b128+(new Array(64)).join(nullbyte)).slice(0,1+((959-(len&511))&511)>>>3)
		p += bigEndianLongWordToString(lhi) + bigEndianLongWordToString(llo)
		var N = (L+p.length)/4, M=0xffffffff, t,X,A,B,C,D,E, k,k4,W,w
		var H0 = 0x67452301, H1 = 0xefcdab89, H2 = 0x98badcfe, H3 = 0x10325476, H4 = 0xc3d2e1f0
		var K0 = 0x5a827999, K1 = 0x6ed9eba1, K2 = 0x8f1bbcdc, K3 = 0xca62c1d6
		for (k = 0; k < N; k += 16)
		{k4 = k<<2
		 W=stringToBigEndianLongWords(k4<L0? msg.substr(k4,64) : k4<L? msg.slice(k4)+p.slice(0,L2) : p.slice(k4-L))
		 for (t = 16; t < 80; w = W[t-3]^W[t-8]^W[t-14]^W[t-16], W[t++] = (w<<1)|(w>>>31) );
		 A=H0, B=H1, C=H2, D=H3, E=H4, t=0
		 for(;t<20;){X=(((A<<5)|(A>>>27))+((B&C)^((~B)&D))   +E+W[t++]+K0)&M,E=D,D=C,C=(B<<30)|(B>>>2),B=A,A=X}
		 for(;t<40;){X=(((A<<5)|(A>>>27))+(B^C^D)            +E+W[t++]+K1)&M,E=D,D=C,C=(B<<30)|(B>>>2),B=A,A=X}
		 for(;t<60;){X=(((A<<5)|(A>>>27))+((B&C)^(B&D)^(C&D))+E+W[t++]+K2)&M,E=D,D=C,C=(B<<30)|(B>>>2),B=A,A=X}
		 for(;t<80;){X=(((A<<5)|(A>>>27))+(B^C^D)            +E+W[t++]+K3)&M,E=D,D=C,C=(B<<30)|(B>>>2),B=A,A=X}
		 H0 = (H0 + A) & M; H1 = (H1 + B) & M; H2 = (H2 + C) & M; H3 = (H3 + D) & M; H4 = (H4 + E) & M;
		}
		return	bigEndianLongWordToString(H0)+bigEndianLongWordToString(H1)
		+	bigEndianLongWordToString(H2)+bigEndianLongWordToString(H3)+bigEndianLongWordToString(H4)
	}
	HRC4.SHA256 = function (msg)
	{	function bigEndianLongWordToString(val)
		{ for(var r="",v=(val<0?val+0x100000000:val),i=0;i<4;r=String.fromByte(v&255)+r,v>>>=8,i++);
		  return r
		}
		function stringToBigEndianLongWords(msg)
		{ for(var r=[],i=0,k=0;i<msg.length;)
		  {for(var v=0,j=0;j<4;v<<=8,v|=(msg.byteAt(i++)&255),j++);r[k++]=v}
		  return r
		}
		var b128 = String.fromByte(128), nullbyte = String.fromByte(0)
		var m = 0x100000000, L=msg.length, L1=L%64, L0=L-L1, L2=64-L1, len=L*8, llo=len%m, lhi=(len-llo)/m
		var p = (b128+(new Array(64)).join(nullbyte)).slice(0,1+((959-(len&511))&511)>>>3)
		p += bigEndianLongWordToString(lhi) + bigEndianLongWordToString(llo)
		var N = (L+p.length)/4, M=0xffffffff, k,k4,blk,W,t
		var K = [0x428a2f98,0x71374491,0xb5c0fbcf,0xe9b5dba5,0x3956c25b,0x59f111f1,0x923f82a4,0xab1c5ed5
			,0xd807aa98,0x12835b01,0x243185be,0x550c7dc3,0x72be5d74,0x80deb1fe,0x9bdc06a7,0xc19bf174
			,0xe49b69c1,0xefbe4786,0x0fc19dc6,0x240ca1cc,0x2de92c6f,0x4a7484aa,0x5cb0a9dc,0x76f988da
			,0x983e5152,0xa831c66d,0xb00327c8,0xbf597fc7,0xc6e00bf3,0xd5a79147,0x06ca6351,0x14292967
			,0x27b70a85,0x2e1b2138,0x4d2c6dfc,0x53380d13,0x650a7354,0x766a0abb,0x81c2c92e,0x92722c85
			,0xa2bfe8a1,0xa81a664b,0xc24b8b70,0xc76c51a3,0xd192e819,0xd6990624,0xf40e3585,0x106aa070
			,0x19a4c116,0x1e376c08,0x2748774c,0x34b0bcb5,0x391c0cb3,0x4ed8aa4a,0x5b9cca4f,0x682e6ff3
			,0x748f82ee,0x78a5636f,0x84c87814,0x8cc70208,0x90befffa,0xa4506ceb,0xbef9a3f7,0xc67178f2
			]
		var H0 = 0x6a09e667, H1 = 0xbb67ae85, H2 = 0x3c6ef372, H3 = 0xa54ff53a
		var H4 = 0x510e527f, H5 = 0x9b05688c, H6 = 0x1f83d9ab, H7 = 0x5be0cd19
		for (k = 0; k < N; k += 16)
		{	k4  = k<<2
			blk = (k4<L0 ? msg.substr(k4,64) : k4<L ? msg.slice(k4)+p.slice(0,L2) : p.slice(k4-L))
			W   = stringToBigEndianLongWords(blk)
			for (t = 16; t < 64; t++)
			{	var x = W[t- 2], sigma1x = ((x>>>17)|(x<<15))^((x>>>19)|(x<<13))^(x>>>10)
				var y = W[t-15], sigma0y = ((y>>> 7)|(y<<25))^((y>>>18)|(y<<14))^(y>>> 3)
				W[t] = (sigma1x + W[t-7] + sigma0y + W[t-16] )&M
			}
			var a = H0, b = H1, c = H2, d = H3, e = H4, f = H5, g = H6, h = H7
			for (t = 0; t < 64; t++)
			{	var SIGMA1e = ((e>>> 6)|(e<<26))^((e>>>11)|(e<<21))^((e>>>25)|(e<< 7))
				var SIGMA0a = ((a>>> 2)|(a<<30))^((a>>>13)|(a<<19))^((a>>>22)|(a<<10))
				var Ch_efg  = (e&f)^((~e)&g)
				var Maj_abc = (a&b)^(a&c)^(b&c)
				var T1 = h + SIGMA1e + Ch_efg + K[t] + W[t]
				var T2 =     SIGMA0a + Maj_abc
				h = g, g = f, f = e, e = (d+T1)&M, d = c, c = b, b = a, a = (T1+T2)&M
			}
			H0=(H0+a)&M,H1=(H1+b)&M,H2=(H2+c)&M,H3=(H3+d)&M,H4=(H4+e)&M,H5=(H5+f)&M,H6=(H6+g)&M,H7=(H7+h)&M
		}
		return	bigEndianLongWordToString(H0)+bigEndianLongWordToString(H1)+bigEndianLongWordToString(H2)
		+	bigEndianLongWordToString(H3)+bigEndianLongWordToString(H4)+bigEndianLongWordToString(H5)
		+	bigEndianLongWordToString(H6)+bigEndianLongWordToString(H7)
	}

// ================================================================================================
// STATIC methods to compute my own varian of a fingerprint, which I've called FP32 and FP64.
// The FP method isused as a general purpose engine, to be considered as an internal method of the
// HRC4 Object.
// Arguments:
//	msg		String with message of which fingerprint should be computed.
// Extra arguments for FP only:
//	length		Size of the desired result.
//	start		Starting byte of the PRNG series to be returned.
//	templength	Size of intermediate key to be used.
//	tempstart	Start position of the intermediate key to be used.
// Returns:
//	String containing the raw binary fingerprint.
// Note:
//	This method uses the HRC4 PRNG with the message as the initialisation key.
//	Then it takes a 'templength' length series of bytes produced by the PRNG and uses that once
//	again as an initialisation key.
//	Then it takes a 'length' bytes substring of the PRNG series starting at the 'start' byte
//	and returns it as a raw binary String.
// ------------------------------------------------------------------------------------------------
	HRC4.FP = function (msg,length,start,templength,tempstart)
	{	var Lr = (arguments.length > 1 ?     length :  32 )	// default result size = 32
		var Sr = (arguments.length > 2 ?     start  :   0 )	// default result skip =  0
		var Lt = (arguments.length > 3 ? templength : 8*Lr)	// default temp   size =  8 * result size
		var St = (arguments.length > 4 ? tempstart  :   0 )	// default temp   skip =  0
		var tmp
		tmp = new HRC4(msg)		; if (St > 0) tmp.random(St)	// get tmp   PRNG and skip if necs
		tmp = new HRC4(tmp.random(Lt))	; if (Sr > 0) tmp.random(Sr)	// get final PRNG and skip if necs
		return tmp.random(Lr)						// return result
	}
	HRC4.FP32 = function (msg)	{return HRC4.FP(msg,32) }
	HRC4.FP64 = function (msg)	{return HRC4.FP(msg,64) }

// ================================================================================================
// STATIC methods to compute a keyed hash of a message.
// Arguments:
//	msg	String with message to be hashed,
//	key	String with key to be used (usually a so called 'nonce'),
// and HMAC also needs:
//	hfun	The hash function to be used,
//	blksz	optional, Number with the blocksize of the hash function, if omitted it
//		defaults to 64, which is the block size in bytes of MD5, SHA1, and SHA256.
//		the FPxx methods have no real block size, but the default of 64 is still used.
// Returns:
//	String containing the raw binary hash value.
// Notes:
//	o The 'hfun' and 'blksz' arguments apply only to the HMAC method,
//	  the HMAC_xxx methods have only the 'msg' and 'key' arguments.
//	o Recommended minimum key length is the size in bytes of the applied hash
//	  (MD5 = 16 bytes, SHA1 is 20 bytes).
//	o In practice, you should use the 'HMAC_xxx' methods and consider the 'HMAC' method
//	  as a generic internal method of the HRC4 Object for use by the 'HMAC_xxx' methods.
// Reference:
//	RFC2104: HMAC: Keyed-Hashing for Message Authentication,
//	http://www.rfc-editor.org/rfc/rfc2104.txt
// ------------------------------------------------------------------------------------------------
	HRC4.HMAC = function (msg, key, hfun, blksz)
	{	var Ki,Ko,B=(arguments.length>3?blksz:64),K=(key.length>B?hfun(key):key)
		K += (new Array(1+B-K.length)).join(String.fromByte(0))
		for (Ki="",i=0; i<B; Ki+=String.fromByte((K.byteAt(i++)^0x36)&255));
		for (Ko="",i=0; i<B; Ko+=String.fromByte((K.byteAt(i++)^0x5c)&255));
		return hfun(Ko+hfun(Ki+msg))
	}
	HRC4.HMAC_MD5    = function (msg, key)	{return HRC4.HMAC(msg,key,HRC4.MD5) }
	HRC4.HMAC_SHA1   = function (msg, key)	{return HRC4.HMAC(msg,key,HRC4.SHA1) }
	HRC4.HMAC_SHA256 = function (msg, key)	{return HRC4.HMAC(msg,key,HRC4.SHA256) }
	HRC4.HMAC_FP32   = function (msg, key)	{return HRC4.HMAC(msg,key,HRC4.FP32) }
	HRC4.HMAC_FP64   = function (msg, key)	{return HRC4.HMAC(msg,key,HRC4.FP64) }

// ================================================================================================
// STATIC methods to encypher or decypher a single message using a given key.
// Arguments:
//	msg		String with message to be encyphered or cypher to be decyphered,
//	key		String with key to be used for encyphering/decyphering.
//	unicode		optional (**crypt only), Boolean whether to use unicode/byte conversion
//			instead of escape/unescape, which reduces the size of the cypher if msg
//			contains LOTS of Unicode characters above 255.
//			if False or omitted then escape/unescape is used.
//	addmarker	see above, in the comment on the Prototype method named 'encrypt'.
// Returns:
//	o String with encyphered or decyphered result.
//	o The 'decypher' and 'decrypt' methods also set the HRC4.lastDecypherIsGood property
//	  to either True or False, indicating whether the decyphering key was correct or not.
// Notes:
//	o the 'encypher' method includes the SHA1 hash of the message in the encyphered result,
//	  the 'decypher' method checks for correct decyphering using this SHA1 hash.
//	o the '**crypt' methods do the same as the '**cypher' methods,
//	  but they do implicit conversion of Unicode characters above 255
//	  and they convert the cypher to/from Base64 notation.
//	o these static methods are for "single shot" mode, i.e. encryption of a single message;
//	  for encrypting multiple consecutive messages it is better to create an HRC4 Object
//	  INSTANCE using the key and then use that instance's methods (i.e. the prototype methods
//	  defined in this file) to encrypt/decrypt the data.
// ------------------------------------------------------------------------------------------------
	HRC4.cypher   = function (msg, key)	{return (new HRC4(key)).cypher  (msg) }
	HRC4.encypher = function (msg, key)	{return (new HRC4(key)).encypher(msg) }
	HRC4.decypher = function (msg, key)	{return (new HRC4(key)).decypher(msg) }

	HRC4.encrypt  = function (msg, key, unicode, addmarker)
	{	return (new HRC4(key)).encrypt (msg, unicode ? true : false, addmarker ? true : false)
	}
	HRC4.decrypt  = function (msg, key, unicode)
	{	return (new HRC4(key)).decrypt (msg, unicode ? true : false)
	}

	HRC4.cypher2   = function (msg, key1, key2)	{return HRC4.new2(key1,key2).cypher  (msg) }
	HRC4.encypher2 = function (msg, key1, key2)	{return HRC4.new2(key1,key2).encypher(msg) }
	HRC4.decypher2 = function (msg, key1, key2)	{return HRC4.new2(key1,key2).decypher(msg) }

	HRC4.encrypt2  = function (msg, key1, key2, unicode, addmarker)
	{	return HRC4.new2(key1,key2).encrypt (msg, unicode ? true : false, addmarker ? true : false)
	}
	HRC4.decrypt2  = function (msg, key1, key2, unicode)
	{	return HRC4.new2(key1,key2).decrypt (msg, unicode ? true : false)
	}


// 2010-02-17 added the 'betterkey' static method:
	HRC4.betterKey = function (key)
	{	var result = ""
		for (var i = 0; i < 8; i++)
		{	if (i%2 != 1)
			{	var hash = HRC4.SHA256(key+result)
				result   = hash + result
			} else
			{	var hash = HRC4.SHA256(result+key)
				result   = result + hash
			}
		}
		return result
	}

// ################################################################################################
// [end of file] hrc4.js -- Copyright (c) 2010 Henk Reints, http://henk-reints.nl
