<?php
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// File:	hrc4.php
//
// Author:	Copyright (c) 2006 Henk Reints, http://henk-reints.nl
//
// Purpose:	Tool for simple ARCFOUR-based data encryption and other binary operations on String Objects.
//
// Original date: 01-Sep-2006
// Last modified: 14-Mar-2007
// Update history: see below
//
// Description:
//	This is a minimally documented translation of hrc4.js from JavaScript to PHP.
//	See hrc4.js for the full documentation.
//
// Usage:
//	require_once ("hrc4.php");
//
// Sample code:
//	$myHRC4object = new HRC4 ($myVerySecretKey);
//	$myEncryptedSecretMessage = $myHRC4object.encrypt ($mySecretMessage);
//	// send $myEncryptedSecretMessage to client
//
//	// receive $myEncryptedSecretMessage from client
//	$myDecryptedSecretMessage = $myHRC4object.decrypt ($myEncryptedSecretMessage);
//	if ($myDecryptedSecretMessage === FALSE)
//	{	echo "Decryption failed!";
//	} else
//	{	echo $myDecryptedSecretMessage;
//	}
//
// Notes:
//	o Not all JS methods are available in this PHP version.
//	o This PHP version does not support Unicode.
//	o Although some functionality is available in later versions of PHP, I have coded some stuff myself
//	  for compatibility with older PHP versions. Everything should work in PHP 4.
//	o The static methods in the JS version are normal functions in this PHP version, all prefixed with "HRC4_".
//	o Some of the functions below are more or less wrappers around built-in PHP functions, they're coded
//	  here just to let this PHP version of HRC4 have the same look and feel as the JS version.
//	o Some extra methods have been added, especially saveSession and restoreSession. Both accept one argument,
//	  being the session variable name (in fact 3 session variables will be used, which are the given name
//	  suffixed with "_I", "_J", and "_S" and prefixed with "HRC4_"). Their usage is:
//
//		$myHRC4Obj = new HRC4 ($key);			// create and initialise an HRC4 object
//		...						// do something with it
//		$myHRC4Obj->saveSession("MYHRC4OBJ");		// save it in the session info (and stop using it!)
//
//	  then, when the PRNG should continue after a new HTTP request:
//
//		$myHRC4Obj = HRC4_restoreSession("MYHRC4OBJ");	// get it back from the session info
//		...						// continue using it where we left in the last request
//		$myHRC4Obj->saveSession("MYHRC4OBJ");		// save it once again
//
// Update history:
//	14-Mar-2007	implemented reshuffle function, as already done in JS on 17-Jan-2007
//	01-Sep-2006	original translation to PHP
//
// DISCLAIMER:
//	HENK REINTS HAS WRITTEN THIS SOFWARE FOR HIS OWN USE, AS A HOBBY, WITHOUT ANY COMMERCIAL INTENTIONS.
//	IT IS PROVIDED AS IS, WITHOUT ANY GUARANTEE FOR CORRECTNESS OR USABILITY. USAGE IS AT YOUR OWN RISK.
//	IN NO WAY SHALL HENK REINTS BE LIABLE FOR ANY NEGATIVE CONSEQUENCES OF USING THIS SOFTWARE.
//
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// =========================================================================
// charAt and byteAt are not in the JS version,
// there some stuff is added to the String Constructor and Prototype instead
// -------------------------------------------------------------------------
	function charAt ($pos, $txt)	{return substr($txt,$pos,1); }
	function byteAt ($pos, $txt)	{return ord(charAt($pos,$txt))&255; }

// =========================================
// define the class, only the dynamic stuff:
// -----------------------------------------
	class HRC4
	{			// I'd prefer to use "private $i,$j,$s;", but PHP 4 seems not to support that
		var $i,$j,$s;	// the internal PRNG status of the Object instance - DO NOT DIRECTLY ACCESS !!!
		// =========================================================================
		// the 'lastDecypherIsGood' property is not implemented in PHP, but instead,
		// the 'decrypt' methods return boolean FALSE on failure, to be used like:
		//	$result = HRC4_decrypt ($msg, $key);
		//	if ($result === FALSE)
		//	{	echo "ERROR";
		//	} else
		//	{	echo $result;
		//	}
		// -------------------------------------------------------------------------
	// ===========================================================================================
	// define the Constructor, if no key is given then the internal status will not be initialised
	// -------------------------------------------------------------------------------------------
		function HRC4 ($key = "", $arcfourMode = false)
		{	if ($key != "") $this->init ($key, $arcfourMode);
		}
	// =============================================================================================
	// next is what's called "prototype methods" in JS, they act on each individual object instance:
	// ---------------------------------------------------------------------------------------------
		function init ($key, $arcfourMode = false)
		{	$L = strlen($key);
			$M = ($arcfourMode ? 256 : max (512, 2*$L) );
			$A = ($arcfourMode ?   0 : 997);
			for ($i = 0; $i < 256; $this->s[$i] = $i, $i++);
			for ($i = 0, $j = 0, $k = 0, $n = 0; $n < $M; $i++, $k++, $n++)
			{	$i &= 255;
				$j = ($j + $this->s[$i] + byteAt($k%$L,$key) ) & 255;
				$t = $this->s[$i];
				$this->s[$i] = $this->s[$j];
				$this->s[$j] = $t;
			}
			for ($i = 0, $j = 0, $n = 0; $n < $A; $n++)
			{	$i = ($i + 1) & 255;
				$j = ($j + $this->s[$i] ) & 255;
				$t = $this->s[$i];
				$this->s[$i] = $this->s[$j];
				$this->s[$j] = $t;
			}
			$this->i = $i;
			$this->j = $j;
		}
		function reshuffle ($key)
		{	$L = strlen($key);
			for ($i = $this->i; $j = $this->j; $k = 0; $k < $L; i++, k++)
			{	$i &= 255;
				$j = ($j + $this->s[$i] + byteAt($k%$L,$key) ) & 255;
				$t = $this->s[$i];
				$this->s[$i] = $this->s[$j];
				$this->s[$j] = $t;
			}
			$this->i = $i;
			$this->j = $j;
		}
		// ===========================================
		// the "dump" method is not implemented in PHP
		// -------------------------------------------
		function random ($cnt = 1)
		{	$result = "";
			for ($i = $this->i, $j = $this->j, $n = 0; $n < $cnt; $n++)
			{	$i = ($i + 1) & 255;
				$j = ($j + $this->s[$i] ) & 255;
				$t = $this->s[$i];
				$this->s[$i] = $this->s[$j];
				$this->s[$j] = $t;
				$result .= chr ($this->s[($this->s[$i]+$this->s[$j])&255] ); 
			}
			$this->i = $i;
			$this->j = $j;
			return $result;
		}
		// =================================================
		// the "randomReal" method is not implemented in PHP
		// -------------------------------------------------
		function randomString ($len)
		{	$D = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
			$result = "";
			for ($i = $this->i, $j = $this->j, $n = 0; $n < $len; $n++)
			{	$i = ($i + 1) & 255;
				$j = ($j + $this->s[$i] ) & 255;
				$t = $this->s[$i];
				$this->s[$i] = $this->s[$j];
				$this->s[$j] = $t;
				$result .= charAt($this->s[($this->s[$i]+$this->s[$j])&255]&0x3f, $D);
			}
			$this->i = $i;
			$this->j = $j;
			return $result;
		}
		function cypher ($msg)
		{	$i = $this->i;
			$j = $this->j;
			$M = strlen($msg);
			$result = "";
			for ($n = 0; $n < $M; $n++)
			{	$i = ($i + 1) & 255;
				$j = ($j + $this->s[$i] ) & 255;
				$t = $this->s[$i];
				$this->s[$i] = $this->s[$j];
				$this->s[$j] = $t;
				$c = byteAt($n,$msg);
				$k = ($this->s[$i] + $this->s[$j])&255;
				$result .= chr($c ^ $this->s[$k]);
			}
			$this->i = $i;
			$this->j = $j;
			return $result;
		}
		function encypher ($msg)
		{	return $this->cypher (HRC4_SHA1($msg) . $msg);
		}
		function decypher ($msg)
		{	$tmp  = $this->cypher ($msg);
			$txt  = substr($tmp,20);
			$good = (HRC4_SHA1($txt) == substr($tmp,0,20) );
			return ($good ? $txt : $good);
		}
		// ===============================================
		// in the methods below, Unicode is not supported:
		// -----------------------------------------------
		function encrypt ($msg)
		{	return HRC4_toBase64($this->encypher(HRC4_escape($msg)));
		}
		function decrypt ($msg)
		{	$tmp = $this->decypher(HRC4_fromBase64($msg));
			if ($tmp !== false) $tmp = HRC4_unescape($tmp);
			return $tmp;
		}
		// ==================================================================
		// saveSession and restoreSession are not available in the JS version
		// ------------------------------------------------------------------
		function saveSession($varnam)
		{	$_SESSION["HRC4_".$varnam."_I"] = $this->i;
			$_SESSION["HRC4_".$varnam."_J"] = $this->j;
			$_SESSION["HRC4_".$varnam."_S"] = $this->s;
		}
		function restoreSession($varnam)
		{	$this->i = $_SESSION["HRC4_".$varnam."_I"];
			$this->j = $_SESSION["HRC4_".$varnam."_J"];
			$this->s = $_SESSION["HRC4_".$varnam."_S"];
		}
	} // end of class HRC4

// ==============================================================================================================
// All methods that are static in JS are defined as normal stand-alone functions below, all prefixed with "HRC4_"
// --------------------------------------------------------------------------------------------------------------
	// =================================================
	// restoreSession is not available in the JS version
	// -------------------------------------------------
	function HRC4_restoreSession($varnam)
	{	$result = new HRC4();	// do not initialise the internal PRNG status
		$result->restoreSession($varnam);
		return $result;
	}
	// ==============================================================================================
	// next escapes only "%", but not "%uXXXX" where X = hex digit, since that's not unescaped by PHP
	// ----------------------------------------------------------------------------------------------
	function HRC4_escape ($msg)
	{	$result = "";
		$L = strlen($msg);
		for ($i = 0; $i < $L; $i++)
		{	$c = charAt($i,$msg);
			if ($c == "%")
			{	$x = substr($msg,$i+1,5);
				if (preg_match("/^u[0-9a-f]{4}$/i",$x) == 0) $c = rawurlencode($c);
			}
			$result .= $c;
		}
		return $result;
	}
	// =============================================================================================
	// the unescape method is not implemented in the JS version, where the built-in unescape is used
	// ---------------------------------------------------------------------------------------------
	function HRC4_unescape ($msg)
	{	return rawurldecode($msg);
	}
	function HRC4_toHex ($msg, $sep = "")
	{	$ss = preg_replace("/[a-f0-9]+/i","",$sep);
		$result = "";
		for ($s = "", $i = 0, $L  = strlen($msg); $i < $L; $s = $ss, $i++)
		{	$result .= $s . substr(dechex(256|byteAt($i,$msg)),1,2);
		}
		return $result;
	}
	function HRC4_fromHex ($msg)
	{	$A = explode(",",preg_replace("/[^a-f0-9,]+/i",",",$msg));
		$result = "";
		for ($i = 0, $M = count($A); $i < $M; $i++)
		{	$L = strlen($A[$i]);
			$H = ($L < 1 ? "00" : ($L&2 == 1 ? "0".$A[$i] : $A[$i] ));
			for ($j = 0, $N = strlen($H); $j < $N; $result .= chr(hexdec(substr($H,$j,2))), $j += 2);
		}
		return $result;
	}
	function HRC4_toBase64 ($msg, $wrap = false, $nopad = false)
	{	$tmp = base64_encode($msg);
		if ($nopad) $tmp = preg_replace("/=+$/","",$tmp);
		$w = ($wrap === true ? 76 : $wrap);
		if (is_int($w))
		{	$res = ""; $L = strlen($tmp); $crlf = chr(13) . chr(10);
			for ($i = 0; $i < $L; $res .= substr($tmp,$i,$w) . crlf, $i += $w);
			return $result;
		} else	return $tmp;
	}
	function HRC4_fromBase64 ($msg)
	{	return base64_decode(preg_replace("/[^a-z0-9\+\/]+/i","",$msg));
	}
	function HRC4_toQuotPr ($msg, $multiline = false)
	{	if ($multiline)
		{	$tmp = explode ("\n", preg_replace ("/\r/", "\n", preg_replace ("/\r\n/", "\n", $msg) ) );
			for ($i = 0, $N = count($tmp); i < $N; $tmp[$i] = HRC4_toQuotPr($tmp[$i]), $i++);
			return implode ("\r\n", $tmp);
		}
		$result = "";
		$softline = "=\r\n";
		for ($i = 0, $M = strlen($msg), $W = 76, $L = 0; $i < $M; )
		{	$c = charAt($i,$msg); $i++; if ($i == $M) $W++;
			if ($c < " " || $c == "=" || $c > "~" || ($c == " " && $i == M) ) $c = "=" . strtoupper(HRC4_toHex($c));
			$L += strlen($c); if ($L < $W) {$result .= $c;} else {$result .= $softline.$c; $L = strlen($c);}
		}
		return $result;
	}
	function HRC4_fromQuotPr ($msg)
	{	$y = preg_replace ("/=\n/", "", preg_replace ("/\r/", "\n", preg_replace ("/\r\n/", "\n", $msg) ) );
		$soft = preg_match ("/=$/", $y); if ($soft > 0) $y = preg_replace ("/=$/", "", $y);
		$result = "";
		for ($i = 0, $L = strlen($y); $i < $L; )
		{	$c = charAt($i,$y); $i++;
			if ($c == "=") {$c = HRC4_fromHex(substr($y,$i,2)); $i += 2;}
			$result .= $c;
		}
		if ($soft == 0) $result .= "\r\n";
		return $result;
	}
	// ====================================================================
	// methods unicodeToBytes and bytesToUnicode are not implemented in PHP
	// --------------------------------------------------------------------
	function HRC4_MD5  ($msg) {return HRC4_fromHex(md5 ($msg)); }
	function HRC4_SHA1 ($msg) {return HRC4_fromHex(sha1($msg)); }
	// ===========================================================
	// SHA256 is not (yet) implemented in this PHP version of HRC4
	// -----------------------------------------------------------
	function HRC4_FP ($msg, $length = 32, $start = 0, $templength = "", $tempstart = 0)
	{	$Lr = $length;						$Sr = $start;
		$Lt = (is_int($templength) ? $templength : 8*$Lr);	$St = $tempstart;
		$tmp = new HRC4($msg);			if ($St > 0) $tmp->random($St);
		$tmp = new HRC4($tmp->random($Lt));	if ($Sr > 0) $tmp->random($Sr);
		return $tmp->random($Lr);
	}
	function HRC4_FP32 ($msg)	{return HRC4_FP($msg,32); }
	function HRC4_FP64 ($msg)	{return HRC4_FP($msg,64); }

	function HRC4_HMAC ($msg, $key, $hfun, $blksz = 64)
	{	$K = (strlen($key) > $blksz ? $hfun($key) : $key);
		for (; strlen($K) < $blksz; $K .= chr(0) );
		for ($Ki = "", $i = 0; $i < $blksz; $Ki .= chr((byteAt($i,$K)^0x36)&255), $i++);
		for ($Ko = "", $i = 0; $i < $blksz; $Ko .= chr((byteAt($i,$K)^0x5c)&255), $i++);
		return $hfun($Ko.$hfun($Ki.$msg));
	}
	function HRC4_HMAC_MD5    ($msg, $key)	{return HRC4_HMAC ($msg, $key, HRC4_MD5   ); }
	function HRC4_HMAC_SHA1   ($msg, $key)	{return HRC4_HMAC ($msg, $key, HRC4_SHA1  ); }
// ##	function HRC4_HMAC_SHA256 ($msg, $key)	{return HRC4_HMAC ($msg, $key, HRC4_SHA256); }	// SHA256 not yet available
	function HRC4_HMAC_FP32   ($msg, $key)	{return HRC4_HMAC ($msg, $key, HRC4_FP32  ); }
	function HRC4_HMAC_FP64   ($msg, $key)	{return HRC4_HMAC ($msg, $key, HRC4_FP64  ); }

	function HRC4_cypher    ($msg, $key)	{$tmp = new HRC4($key); return $tmp->cypher  ($msg); }
	function HRC4_encypher  ($msg, $key)	{$tmp = new HRC4($key); return $tmp->encypher($msg); }
	function HRC4_decypher  ($msg, $key)	{$tmp = new HRC4($key); return $tmp->decypher($msg); }
	function HRC4_encrypt   ($msg, $key)	{$tmp = new HRC4($key); return $tmp->encrypt ($msg); }
	function HRC4_decrypt   ($msg, $key)	{$tmp = new HRC4($key); return $tmp->decrypt ($msg); }

// ################################################################################################
// [end of file] hrc4.php -- Copyright (c) 2006 Henk Reints, http://henk-reints.nl
?>
